# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import io
import json
from typing import Dict, List, Optional

from .. import app, convert, experiment_api, experiment_info
from ..types import JSONEncodable
from . import chain, state


def start_chain(
    inputs: Dict[str, JSONEncodable],
    api_key: Optional[str] = None,
    workspace: Optional[str] = None,
    project: Optional[str] = None,
    metadata: Optional[Dict[str, Dict[str, JSONEncodable]]] = None,
    tags: Optional[List[str]] = None,
) -> None:
    """
    Creates global Chain object that tracks created Spans.
    Args:
        inputs: Dict[str, JSONEncodable] (required) chain inputs.
        workspace: str (optional) comet workspace to use for logging.
        project: str (optional) project name to create in comet workspace.
        tags: List[str] (optional), user-defined tags attached to a prompt call.
        api_key: str (optional) comet API key.
        metadata: Dict[str, Dict[str, JSONEncodable]] (optional) user-defined
            dictionary with additional metadata to the call.
        tags: List[str] (optional) user-defined tags attached to the chain
    """

    MESSAGE = """
    CometLLM requires an API key. Please provide it as the
    api_key argument to comet_llm.start_chain or as an environment
    variable named COMET_API_KEY
    """

    experiment_info_ = experiment_info.get(
        api_key,
        workspace,
        project,
        api_key_not_found_message=MESSAGE,
    )
    global_chain = chain.Chain(
        inputs=inputs,
        metadata=metadata,
        experiment_info=experiment_info_,
        tags=tags,
    )
    state.set_global_chain(global_chain)


def end_chain(
    outputs: Dict[str, JSONEncodable],
    metadata: Optional[Dict[str, JSONEncodable]] = None,
) -> None:
    """
    Commits global chain and logs the result to Comet.
    Args:
        outputs: Dict[str, JSONEncodable] (required) chain outputs.
        metadata: Dict[str, Dict[str, JSONEncodable]] (optional) user-defined
            dictionary with additional metadata to the call. This metadata
            will be deep merged with the metadata passed to start_chain if
            it was provided.
        tags: List[str] (optional) user-defined tags attached to the chain
    """
    global_chain = state.get_global_chain()
    global_chain.set_outputs(outputs=outputs, metadata=metadata)
    global_chain_data = global_chain.as_dict()

    experiment_info_ = global_chain.experiment_info
    experiment_api_ = experiment_api.ExperimentAPI(
        api_key=experiment_info_.api_key,
        workspace=experiment_info_.workspace,
        project_name=experiment_info_.project_name,
    )

    if global_chain.tags is not None:
        experiment_api_.log_tags(global_chain.tags)

    experiment_api_.log_asset_with_io(
        name="comet_llm_data.json",
        file=io.StringIO(json.dumps(global_chain_data)),
        asset_type="llm_data",
    )

    experiment_api_.log_metric(
        name="chain_duration", value=global_chain_data["chain_duration"]
    )

    parameters = convert.chain_metadata_to_flat_parameters(
        global_chain_data["metadata"]
    )
    for name, value in parameters.items():
        experiment_api_.log_parameter(name, value)

    app.SUMMARY.add_log(experiment_api_.project_link, "chain")



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

from typing import TYPE_CHECKING, Optional

from .. import exceptions

if TYPE_CHECKING:  # pragma: no cover
    from .. import experiment_info
    from . import chain


class State:
    def __init__(self) -> None:
        self._id: int = 0
        self._chain: Optional["chain.Chain"] = None

        self._experiment_info: Optional["experiment_info.ExperimentInfo"] = None

    @property
    def chain(self) -> "chain.Chain":
        if self._chain is None:
            raise exceptions.CometLLMException(
                "Global chain is not initialized. Initialize it with `comet_llm.start_chain(...)`"
            )

        return self._chain

    @chain.setter
    def chain(self, new_chain: "chain.Chain") -> None:
        self._chain = new_chain

    def new_id(self) -> int:
        self._id += 1
        return self._id


APP_STATE = State()


def get_global_chain() -> "chain.Chain":
    return APP_STATE.chain


def set_global_chain(new_chain: "chain.Chain") -> None:
    APP_STATE.chain = new_chain


def get_new_id() -> int:
    return APP_STATE.new_id()



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import collections
from typing import DefaultDict, Dict


class LogsRegistry:
    def __init__(self) -> None:
        self._registry: DefaultDict[str, int] = collections.defaultdict(lambda: 0)

    def register_log(self, project_url: str) -> None:
        self._registry[project_url] += 1

    def as_dict(self) -> Dict[str, int]:
        return self._registry.copy()

    def empty(self) -> bool:
        return len(self._registry) == 0



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

from typing import Optional

from . import datetimes, exceptions


def timestamp(timestamp: Optional[float]) -> float:
    if timestamp is None:
        return datetimes.local_timestamp()

    if not datetimes.is_valid_timestamp_seconds(timestamp):
        raise exceptions.CometLLMException(
            "Invalid timestamp: {timestamp}. Timestamp must be in seconds if specified."
        )

    return timestamp * 1000



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import sys
from importlib import machinery
from types import ModuleType
from typing import List, Optional

from . import module_loader, registry


class CometFinder:
    def __init__(self, extensions_registry: registry.Registry) -> None:
        self._registry = extensions_registry
        self._pathfinder = machinery.PathFinder()

    def hook_into_import_system(self) -> None:
        if self not in sys.meta_path:
            sys.meta_path.insert(0, self)  # type: ignore

    def find_spec(
        self, fullname: str, path: Optional[List[str]], target: Optional[ModuleType]
    ) -> Optional[machinery.ModuleSpec]:
        if fullname not in self._registry.module_names:
            return None

        original_spec = self._pathfinder.find_spec(fullname, path, target)

        if original_spec is None:
            return None

        return self._wrap_spec_loader(fullname, original_spec)

    def _wrap_spec_loader(
        self, fullname: str, spec: machinery.ModuleSpec
    ) -> machinery.ModuleSpec:
        module_extension = self._registry.get_extension(fullname)
        spec.loader = module_loader.CometModuleLoader(fullname, spec.loader, module_extension)  # type: ignore
        return spec



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************


import functools
import urllib.parse
from typing import Any, Callable, List

import requests  # type: ignore

from .. import config, exceptions


def wrap(check_on_prem: bool = False) -> Callable:
    def inner_wrap(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Any:  # type: ignore
            try:
                return func(*args, **kwargs)
            except requests.RequestException as exception:
                exception_args: List[Any] = []

                if check_on_prem:
                    comet_url = config.comet_url()
                    if _is_on_prem(comet_url):
                        exception_args.append(
                            f"Failed to send prompt to your Comet installation at "
                            f"{comet_url}. Check that your Comet "
                            f"installation is up-to-date and check the traceback for more details."
                        )

                raise exceptions.CometLLMException(*exception_args) from exception

        return wrapper

    return inner_wrap


def _is_on_prem(url: str) -> bool:
    parsed = urllib.parse.urlparse(url)
    root = f"{parsed.scheme}://{parsed.hostname}/"
    return root != "https://www.comet.com/"



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

from typing import Dict, Optional

from .. import datetimes
from ..types import JSONEncodable
from . import state


class Span:
    """
    A single unit of Chain that has its own
    context. Spans can be nested, in that case inner one
    exist in the context of the outer(parent) one.
    Outer Span is considered to be a parent for an inner one.
    """

    def __init__(
        self,
        inputs: JSONEncodable,
        category: str,
        name: Optional[str] = None,
        metadata: Optional[Dict[str, JSONEncodable]] = None,
    ):
        """
        Args:
            inputs: JSONEncodable (required), span inputs.
            category: str (required), span category.
            name: str (optional), span name. If not set will be
                generated automatically.
            metadata: Dict[str, JSONEncodable] (optional), span metadata.
        """
        self._inputs = inputs
        self._category = category
        self._metadata = metadata if metadata is not None else {}
        self._outputs: Optional[Dict[str, JSONEncodable]] = None
        self._id = state.get_new_id()

        self._connect_to_chain()
        self._name = (
            name if name is not None else self._chain.generate_node_name(category)
        )

        self._timer = datetimes.Timer()

    def _connect_to_chain(self) -> None:
        chain = state.get_global_chain()
        chain.track_node(self)
        self._context = chain.context.current()
        self._chain = chain

    @property
    def id(self) -> int:  # pragma: no cover
        return self._id

    @property
    def name(self) -> str:  # pragma: no cover
        return self._name

    def __enter__(self) -> "Span":
        self._timer.start()
        self._chain.context.add(self.id)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        self._timer.stop()
        self._chain.context.pop()

    def set_outputs(
        self,
        outputs: Dict[str, JSONEncodable],
        metadata: Optional[Dict[str, JSONEncodable]] = None,
    ) -> None:
        """
        Sets outputs to span object.
        Args:
            outputs: Dict[str, JSONEncodable] (required), outputs
            metadata: Dict[str, JSONEncodable] (optional), span metadata.
                If metadata dictionary was passed to __init__ method,
                it will be updated.
        """
        self._outputs = outputs
        if metadata is not None:
            self._metadata.update(metadata)

    def as_dict(self) -> Dict[str, JSONEncodable]:
        inputs = self._inputs
        outputs = self._outputs

        inputs = inputs if isinstance(inputs, dict) else {"input": inputs}
        outputs = outputs if isinstance(outputs, dict) else {"output": outputs}

        return {
            "id": self._id,
            "category": self._category,
            "name": self._name,
            "inputs": inputs,
            "outputs": outputs,
            "duration": self._timer.duration,
            "start_timestamp": self._timer.start_timestamp,
            "end_timestamp": self._timer.end_timestamp,
            "parent_ids": self._context,
            "metadata": self._metadata,
        }



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import functools
import urllib.parse
from typing import IO, List, Optional

import requests  # type: ignore

from .. import config
from ..types import JSONEncodable
from . import request_exception_wrapper

ResponseContent = JSONEncodable


class CometAPIClient:
    def __init__(self, api_key: str, comet_url: str):
        self._headers = {"Authorization": api_key}
        self._comet_url = comet_url

    def create_experiment(
        self,
        type_: str,
        workspace: Optional[str],
        project: Optional[str],
    ) -> ResponseContent:
        return self._request(
            "POST",
            "/api/rest/v2/write/experiment/create",
            json={"workspaceName": workspace, "projectName": project, "type": type_},
        )

    def log_experiment_parameter(
        self, experiment_key: str, name: str, value: JSONEncodable
    ) -> ResponseContent:
        return self._request(
            "POST",
            "/api/rest/v2/write/experiment/parameter",
            json={
                "experimentKey": experiment_key,
                "parameterName": name,
                "parameterValue": value,
            },
        )

    def log_experiment_metric(
        self, experiment_key: str, name: str, value: JSONEncodable
    ) -> ResponseContent:
        return self._request(
            "POST",
            "/api/rest/v2/write/experiment/metric",
            json={
                "experimentKey": experiment_key,
                "metricName": name,
                "metricValue": value,
            },
        )

    def log_experiment_asset_with_io(
        self,
        experiment_key: str,
        name: str,
        file: IO,
        asset_type: str,
        extension: Optional[str] = None,
    ) -> ResponseContent:
        extension = name.split(".")[-1] if extension is None else extension

        return self._request(
            "POST",
            "/api/rest/v2/write/experiment/upload-asset",
            params={
                "experimentKey": experiment_key,
                "fileName": name,
                "extension": extension,
                "type": asset_type,
            },
            files={"file": file},
        )

    def log_experiment_tags(
        self, experiment_key: str, tags: List[str]
    ) -> ResponseContent:
        return self._request(
            "POST",
            "/api/rest/v2/write/experiment/tags",
            json={"experimentKey": experiment_key, "addedTags": tags},
        )

    def _request(self, method: str, path: str, *args, **kwargs) -> ResponseContent:  # type: ignore
        url = urllib.parse.urljoin(self._comet_url, path)
        response = requests.request(method, url, headers=self._headers, *args, **kwargs)
        response.raise_for_status()
        return response.json()


@functools.lru_cache(maxsize=1)
def get(api_key: str) -> CometAPIClient:
    comet_url = config.comet_url()
    comet_api_client = CometAPIClient(api_key, comet_url)

    return comet_api_client



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import logging
from typing import Any, Callable, Dict, List, Tuple, Union

from . import validate
from .types import AfterCallback, AfterExceptionCallback, BeforeCallback

LOGGER = logging.getLogger(__name__)

Args = Union[Tuple[Any, ...], List[Any]]
ArgsKwargs = Tuple[Args, Dict[str, Any]]


def run_before(  # type: ignore
    callbacks: List[BeforeCallback], original: Callable, *args, **kwargs
) -> ArgsKwargs:
    for callback in callbacks:
        try:
            callback_return = callback(original, *args, **kwargs)

            if validate.args_kwargs(callback_return):
                LOGGER.debug("New args %r", callback_return)
                args, kwargs = callback_return
        except Exception:
            LOGGER.debug(
                "Exception calling before callback %r", callback, exc_info=True
            )

    return args, kwargs


def run_after(  # type: ignore
    callbacks: List[AfterCallback],
    original: Callable,
    return_value: Any,
    *args,
    **kwargs
) -> Any:
    for callback in callbacks:
        try:
            new_return_value = callback(original, return_value, *args, **kwargs)
            if new_return_value is not None:
                return_value = new_return_value
        except Exception:
            LOGGER.debug("Exception calling after callback %r", callback, exc_info=True)

    return return_value


def run_after_exception(  # type: ignore
    callbacks: List[AfterExceptionCallback],
    original: Callable,
    exception: Exception,
    *args,
    **kwargs
) -> None:
    for callback in callbacks:
        try:
            callback(original, exception, *args, **kwargs)
        except Exception:
            LOGGER.debug(
                "Exception calling after-exception callback %r", callback, exc_info=True
            )



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import io
import json
from typing import Dict, List, Optional, Union

from . import app, convert, experiment_api, experiment_info, preprocess
from .chains import version


def log_prompt(
    prompt: str,
    output: str,
    workspace: Optional[str] = None,
    project: Optional[str] = None,
    tags: Optional[List[str]] = None,
    api_key: Optional[str] = None,
    prompt_template: Optional[str] = None,
    prompt_template_variables: Optional[
        Dict[str, Union[str, bool, float, None]]
    ] = None,
    metadata: Optional[Dict[str, Union[str, bool, float, None]]] = None,
    timestamp: Optional[float] = None,
    duration: Optional[float] = None,
) -> None:
    """
    Logs a single prompt and output to Comet platform.

    Args:
        prompt: str (required) input prompt to LLM.
        output: str (required), output from LLM.
        workspace: str (optional) comet workspace to use for logging.
        project: str (optional) project name to create in comet workspace.
        tags: List[str] (optional), user-defined tags attached to a prompt call.
        api_key: str (optional) comet API key.
        prompt_template: str (optional) user-defined template used for creating a prompt.
        prompt_template_variables: Dict[str, str] (optional) dictionary with data used
            in prompt_template to build a prompt.
        metadata: Dict[str, Union[str, bool, float, None]] (optional) user-defined
            dictionary with additional metadata to the call.
        timestamp: float (optional) timestamp of prompt call in seconds
        duration: float (optional) duration of prompt call

    Example:

    ```python
    log_prompt(
        prompt="Answer the question and if the question can't be answered, say \"I don't know\"\n\n---\n\nQuestion: What is your name?\nAnswer:",
        metadata={
            "input.type": "completions",
            "input.model": "text-davinci-003",
            "input.provider": "openai",
            "output.index": 0,
            "output.logprobs": None,
            "output.finish_reason": "length",
            "usage.prompt_tokens": 5,
            "usage.completion_tokens": 7,
            "usage.total_tokens": 12,
        },
        prompt_template="Answer the question and if the question can't be answered, say \"I don't know\"\n\n---\n\nQuestion: {{question}}?\nAnswer:",
        prompt_template_variables={"question": "What is your name?"},
        output=" My name is [your name].",
        duration=16.598,
    )

    ```

    Returns: None.
    """
    LOG_PROMPT_API_KEY_NOT_FOUND_MESSAGE = """
    CometLLM requires an API key. Please provide it as the
    api_key argument to log_prompt or as an environment
    variable named COMET_API_KEY
    """
    timestamp = preprocess.timestamp(timestamp)

    info = experiment_info.get(
        api_key,
        workspace,
        project,
        api_key_not_found_message=LOG_PROMPT_API_KEY_NOT_FOUND_MESSAGE,
    )
    experiment_api_ = experiment_api.ExperimentAPI(
        api_key=info.api_key, workspace=info.workspace, project_name=info.project_name
    )

    call_data = convert.call_data_to_dict(
        prompt=prompt,
        outputs=output,
        metadata=metadata,
        prompt_template=prompt_template,
        prompt_template_variables=prompt_template_variables,
        start_timestamp=timestamp,
        end_timestamp=timestamp,
        duration=duration,
    )

    asset_data = {
        "version": version.ASSET_FORMAT_VERSION,
        "chain_nodes": [call_data],
        "chain_inputs": {
            "final_prompt": prompt,
            "prompt_template": prompt_template,
            "prompt_template_variables": prompt_template_variables,
        },
        "chain_outputs": {"output": output},
        "category": "single_prompt",
        "metadata": metadata,
        "start_timestamp": timestamp,
        "end_timestamp": timestamp,
        "chain_duration": duration,
    }

    experiment_api_.log_asset_with_io(
        name="comet_llm_data.json",
        file=io.StringIO(json.dumps(asset_data)),
        asset_type="llm_data",
    )

    if tags is not None:
        experiment_api_.log_tags(tags)

    if duration is not None:
        experiment_api_.log_metric("chain_duration", duration)

    parameters = convert.chain_metadata_to_flat_parameters(metadata)

    for name, value in parameters.items():
        experiment_api_.log_parameter(name, value)

    app.SUMMARY.add_log(experiment_api_.project_link, "prompt")



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

import importlib.abc
from types import ModuleType
from typing import TYPE_CHECKING, Optional

from . import module_extension, patcher

if TYPE_CHECKING:  # pragma: no cover
    from importlib import machinery


class CometModuleLoader(importlib.abc.Loader):
    def __init__(
        self,
        module_name: str,
        original_loader: importlib.abc.Loader,
        module_extension: module_extension.ModuleExtension,
    ) -> None:
        self._module_name = module_name
        self._original_loader = original_loader
        self._module_extension = module_extension

    def create_module(self, spec: "machinery.ModuleSpec") -> Optional[ModuleType]:
        if hasattr(self._original_loader, "create_module"):
            return self._original_loader.create_module(spec)

        LET_PYTHON_HANDLE_THIS = None
        return LET_PYTHON_HANDLE_THIS

    def exec_module(self, module: ModuleType) -> None:
        if hasattr(self._original_loader, "exec_module"):
            self._original_loader.exec_module(module)
        else:
            module = self._original_loader.load_module(self._module_name)

        patcher.patch(module, self._module_extension)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************

from typing import Dict

from . import callable_extenders


class ModuleExtension:
    def __init__(self) -> None:
        self._callables_extenders: Dict[str, callable_extenders.CallableExtenders] = {}

    def extenders(self, callable_name: str) -> callable_extenders.CallableExtenders:
        if callable_name not in self._callables_extenders:
            self._callables_extenders[callable_name] = callable_extenders.get()

        return self._callables_extenders[callable_name]

    def items(self):  # type: ignore
        return self._callables_extenders.items()
